#/usr/bin/env python3.4
#
# Filename: ecp_pki_ref_client.py
# Pre-requisites: Python 3.4, lxml 3.4.2 ECP enabled SP and IdP. Tested with Shibboleth SP 2.5.3 and Shibboleth IdP 2.4.3.
#
# Function: To establish a session with an ECP enabled SAML service provider, SP, using a Shibboleth IdP via ECP.
#	This script uses a public key infrastructure, PKI, certificate for authentication to the IdP. It utilizes cookies
#	and injesting a file with a list of "\n" separated URLs protected by the SP. This allows a crawler or other
#	automated process to create the initial session, then reuse the session to access other protected resources. A file or 
#	a single URL can be passed into this python program.
# Example commands: python3.4 ecp_pki_ref_client.py --file "listofurls.txt" --debug=0
#	python3.4 ecp_pki_ref_client.py --url "https://<hostname>/secure/" --debug=1
#
# Modified by R. Eastman, eastmanrg<at>gmail<dot>com
# Orignial Script: https://wiki.shibboleth.net/confluence/download/attachments/4358416/ecp.py?api=v2
# Date: April 2015
# Version 1.0.0
#

import os
import sys
import stat
import urllib.request
import http.cookiejar
import re
import getpass
import ssl
import http

from optparse import OptionParser
from lxml import etree
from copy import deepcopy

# IDP ECP Endpoint
idp_endpoint = "https://<IDP_HOSTNAME>/SAML2/SOAP/ECP"

# Turn on http debugging by switching to 1
http.client.HTTPConnection.debuglevel = 0

# Cookie file the script saves and loads from
cookieFile = "ssoecpcookie.lwp"

# This is your PKi Certificate in .pem format. You can use just the cert and key seperately or concatenated.
client_cert = "client_cert.pem"
private_key = "client_key.pem"

# This is your certificate bundle
cabundle = "ca_bundle.pem"
#############################################################################################################
# This is an extension to the HTTPSConnection used to connect to the non-pki enabled Service Provider
class HTTPSClientAuthHandler(urllib.request.HTTPSHandler):
	def _init_(self):
		urllib.request.HTTPSHandler._init_(self)
	def https_open(self, req):
		return self.do_open(self.getConnection, req)
	def getConnection(self, host, timeout=300):
		context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
		context.set_default_verify_paths()
		conext.check_hostname = False
		return http.client.HTTPSConnection(host, context=context)
#############################################################################################################
# This is an extension to the HTTPSConnection that allows the use of the 
# PKI client certificates for authentication to the IdP.
class HTTPSIdPClientAuthHandler(urllib.request.HTTPSHandler):
	def _init_(self):
		urllib.request.HTTPSHandler._init_(self)
	def https_open(self, req):
		return self.do_open(self.getConnection, req)
	def getConnection(self, host, timeout=300):
		context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
		context.load_verify_locations(cabundle)
		context.set_default_verify_paths()
		context.load_cert_chain(certfile=client_cert, keyfile=private_key)
		context.verify_mode = ssl.CERT_OPTIONAL
		conext.check_hostname = False
		return http.client.HTTPSConnection(host, context=context)
#############################################################################################################
# This is the core of the program.
# Given an IdP endpoint for ECP, the desired target on the SP tis function manages
#	an ECP exchange with the SP and the IdP and prints the contents of the target to stdout
#	after establishing a session with the SP. It can be configured to save the output to disk.
def get(idp_endpoint, sp_target, debug, lines):
	# Create a cookie jar and cookie handler
	cookie_jar = http.cookiejar.LWPCookieJar(filename=cookieFile)
	if os.path.isfile(cookieFile)
		cookie_jar.load(cookieFile)
	cookie_handler = urllib.request.HTTPCookieProcessor(cookie_jar)

	# Create the base opener object
	opener = urllib.request.build_opener(cookie_handler, HTTPSClientAuthHandler())
	urllib.request.install_opener(opener)

    # Headers needed to indicate to the SP an ECP request
	headers = {
		'Accept' : 'text/html; application/vnd.paos+xml',
		'PAOS'   : 'ver="urn:liberty:paos:2003-08";"urn:oasis:names:tc:SAML:2.0:profiles:SSO:ecp"'
		}
	# Request target from SP 
	request = urllib.request.Request(url=sp_target,headers=headers)

	try:
		response = opener.open(request)
	except Exception as e:
		print ("First request to SP "+sp_target+" failed: %s" % e)
		sys.exit(1)
	# Convert the SP resonse from string to etree Element object
	sp_response = etree.XML(response.read())
	if debug: 
		print
		print ("###### BEGIN SP RESPONSE")
		print
		print (etree.tostring(sp_response))
		print
		print ("###### END SP RESPONSE")
		print

	# Parse out the relay state element from the SP so that it can
	# be included later in the response to the SP
	namespaces = {
		'ecp' : 'urn:oasis:names:tc:SAML:2.0:profiles:SSO:ecp',
		'S'   : 'http://schemas.xmlsoap.org/soap/envelope/',
		'paos': 'urn:liberty:paos:2003-08'
		}

	try:
		relay_state = sp_response.xpath("//ecp:RelayState", namespaces=namespaces)[0]
	except Exception as e:
		print ("Unable to parse relay state element from SP response: %s" % e)
		sys.exit(1)

	if debug: 
		print
		print ("###### BEGIN RELAY STATE ELEMENT")
		print
		print (etree.tostring(relay_state))
		print
		print ("###### END RELAY STATE ELEMENT")
		print

	# Parse out the responseConsumerURL attribute so that it can
	# later be compared with the assertionConsumerURL sent by the IdP
	try:
		response_consumer_url = sp_response.xpath("/S:Envelope/S:Header/paos:Request/@responseConsumerURL", namespaces=namespaces)[0]
	except Exception as e:
		print ("Unable to parse responseConsumerURL attribute from SP response: %s" % e)
		sys.exit(1)

	if debug: 
		print
		print ("###### BEGIN RESPONSE CONSUMER URL")
		print
		print (etree.tostring(relay_state))
		print
		print ("###### END RESPONSE CONSUMER URL")
		print

	# Make a deep copy of the SP response and then remove the header
	# in order to create the package for the IdP
	idp_request = deepcopy(sp_response)
	header = idp_request[0]
	idp_request.remove(header)

	if debug: 
		print
		print ("###### BEGIN IDP REQUEST")
		print
		print (etree.tostring(idp_request))
		print
		print ("###### END IDP REQUEST")
		print

	# Create the base opener object with the IdP https handler
	opener = urllib.request.build_opener(cookie_handler, HTTPSIdPClientAuthHandler())
	urllib.request.install_opener(opener)
	
	# POST the request to the IdP 
	request = urllib.request.Request(idp_endpoint, data=etree.tostring(idp_request))
	request.get_method = lambda: 'POST'
	request.add_header('Content-Type', 'text/html')

	try:
		response = opener.open(request)
	except Exception as e:
		print ("Request to IdP failed: %s" % e)
		sys.exit(1)

	idp_response = etree.XML(response.read())
	if debug: 
		print
		print ("###### BEGIN IDP RESPONSE")
		print
		print (etree.tostring(idp_response))
		print
		print ("###### END IDP RESPONSE")
		print

	try:
		assertion_consumer_service = idp_response.xpath("/S:Envelope/S:Header/ecp:Response/@AssertionConsumerServiceURL", namespaces=namespaces)[0]
	except Exception as e:
		print ("Error parsing assertionConsumerService attribute from IdP response: %s" % e)
		sys.exit(1)

	if debug: 
		print
		print ("###### BEGIN ASSERTION CONSUMER SERVICE URL")
		print
		print (etree.tostring(idp_response))
		print
		print ("###### END ASSERTION CONSUMER SERVICE URL")
		print

	# if the assertionConsumerService attribute from the IdP 
	# does not match the responseConsumerURL from the SP
	# we cannot trust this exchange so send SOAP 1.1 fault
	# to the SP and exit
	if assertion_consumer_service != response_consumer_url:
		print ("ERROR: assertionConsumerServiceURL %s does not" % assertion_consumer_service)
		print ("match responseConsumerURL %s" % response_consumer_url)
		print ("")
		print ("sending SOAP fault to SP")
        
		soap_fault = """
			<S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
			   <S:Body>
				 <S:Fault>
					<faultcode>S:Server</faultcode>
					<faultstring>responseConsumerURL from SP and assertionConsumerServiceURL from IdP do not match</faultstring>
				 </S:Fault>
			   </S:Body>
			</S:Envelope>
			"""

		headers = {
					'Content-Type' : 'application/vnd.paos+xml',
					}
        
		request = urllib.request.Request(url=response_consumer_url, data=soap_fault, headers=headers)
		request.get_method = lambda: 'POST'

		# POST the SOAP 1.1 fault to the SP and ignore any return 
		try:
			response = opener.open(request)
		except Exception as e:
			print ("Error SOAP 1.1 fault: %s" % e)
			pass
			sys.exit(1)

	# Make a deep cop of the IdP response and replace its
	# header contents with the relay state initially sent by
	# the SP
	sp_package = deepcopy(idp_response)
	sp_package[0][0] = relay_state 

	if debug: 
		print
		print ("###### BEGIN PACKAGE TO SEND TO SP")
		print
		print (etree.tostring(sp_package))
		print
		print ("###### END PACKAGE TO SEND TO SP")
		print


	headers = {
		'Content-Type' : 'application/vnd.paos+xml',
		}
	# Create the base opener object
	opener = urllib.request.build_opener(cookie_handler, HTTPSClientAuthHandler())
	urllib.request.install_opener(opener)
	
	# POST the package to the SP
	request = urllib.request.Request(url=assertion_consumer_service, data=etree.tostring(sp_package), headers=headers)
	request.get_method = lambda: 'POST'

	try:
		response = opener.open(request)
	except Exception as e:
		print ("Error POSTing package to SP: %s" % e)
		sys.exit(1)
	# Save the cooke to the file system
	cookie_jar.save(ignore_discard=True, ignore_expires=True)

	# We ignore the response from the SP here and rely on the 
	# opener() instance and the cookie jar to get the cookies
	# we need as they are sent from the SP in order to make the
	# final request

	# We use the existing established session to request the original target
	# from the SP. Multiple calls could be done here using the established session
	# with the SP, and the cookies are saved to the file system
	# to be used with later processes.

	url_response = ""
	for urls in lines:
		if urls != '' and urls != '\n':
			request = urllib.request.Request(url=urls)
			try:
				response = opener.open(request)
				# Save the cookies to the file system
				cookie_jar.save(ignore_discard=True, ignore_expires=True)
				sp_response = response.read()
				url_response += sp_response.decode('ascii')
			except Exception as e:
				print("SP Response using the url_file failed: %s" %e)
	return url_response
#############################################################################################################
def main():
	url_file = ""
	url_get = ""
	parser = OptionParser()
	parser.add_option("--url", action="store", type="string", dest="url", help="Pass in a single URL.")
	parser.add_option("--file", action="store", type="string", dest="file",help="Pass in a file with a newline separated list of URLs.")
	parser.add_option("--debug", action="store", type="int", dest="debug",help="Pass a 1 for true and a 0 for false")
	(options, args) = parser.parse_args()
	if options.url != None:
		lines = []
		lines.insert(0, options.url)
		sp_target = options.url
		url_get = get(idp_endpoint, sp_target, options.debug, lines)
		print (url_get)
		sys.exit(1)
	if options.filename != None:
		url_file = options.filename
		if os.path.exists(url_file):
			lines = []
			file = open(url_file)
			lines = file.read().split('\n')
			file.close()
			sp_target = lines[0]
		else:
			print("File "+url_file+" does not exist.")
			sys.exit(1)
	url_get = get(idp_endpoint, sp_target, options.debug, lines)
	print(url_get)
	sys.exit(1)
#############################################################################################################
if __name__ == "__main__":
	main()

